#############################################################################
#
# Package: The Canon Package
#
# Language: Maple 6, 7, 8, 9.
#
# Description: The Canon Package puts tensor expressions (or expressions with 
#              indexed objects) into the canonical form with respect to index 
#              ordering. The indexed objects can obey generic permutation 
#              symmetries and can have dummy indices.
#              The tensor expression must be in the explicit index notation,
#              and the product must be commutative (this is easily extended 
#              to anticommutative or non-commutative case - it was not done 
#              because Maple does not have a robust non-commutative product).
#              The metric tensor associated with the dummy indices can be
#              symmetric (tensor calculus), antisymmetric (spinor calculus),
#              or can have no symmetry (affine tensor calculus).
#              We are using the name "tensor" meaning "indexed object" in 
#              general.
#
# Developers: Leon R. U. Manssur and Renato Portugal
#             Laboratorio Nacional de Computacao Cientifica (LNCC)
#
# Web page: http://www.cbpf.br/~portugal/Canon.html
#
# Contact and bug report: Portugal@Lncc.Br
#
#############################################################################
#
# COPYRIGHT NOTICE
#
# The Canon package has all rights reserved to the developers. The package 
# and documentation are provided as is, without guarantee of support or 
# maintenance. We are not liable for any damages resulting in any way from 
# their use.
# 
# Everyone is granted permission to copy, modify and redistribute the Canon
# package and its documentation, provided that, (1) all copies contain this
# notice in the main program file and in the documentation, (2) no charge 
# is made for this software or works derived from it.
#
#############################################################################


`Canon/version` := proc()
options `Copyright 2003 by LRU Manssur and R Portugal. All rights reserved`;
  "Canon version 1.1, Jun, 16, 2003, Maple 6,7,8,9"
end proc:


#
# Structure of this file
#
# 1. INITIALIZATION 
#
# 2. USER PROCEDURES
#
# 3. INTERNAL GROUP PROCEDURES
#
# 4. TENSOR PROCEDURES
#
# 5. TYPES (tensor)
#
# 6. INTERFACE
#
# 7. GENERAL TOOLS
#



#############################################################################
#############################################################################
#
# 1. INITIALIZATION


# Global Variables 
# Convention: The name of all global variables starts with "Canon"
# Globals have the drawback with respect to using option remember.

#
# CanonMetricSymmetry - symmetry of the metric
#
CanonMetricSymmetry := 1:  # Allowable values = {1,0,-1} 

#
# CanonTable - table of the symmetries of the tensors defined
#                      by CanonDefine.
#
CanonTable := table():

#
# CanonTensor is used as the name of internal tables. No 
# assingment is performed.
#
protect(CanonTensor):

#
# CanonDummyIndices and CanonFreeIndices are used as global
# variables for communication among tensor procedures. They are cleared
# at the end of each calling.
#

#
# _FREE is a inert function used in internal procedures. It does not
# appear in outputs.
#
protect('_FREE'):

protect('_SCALAR');

#
# The symbols `` and ` ` are used in the output (interface)
#
protect(``):
protect(` `):

# Macros 

#macro(printtensor=CanonPrint):



#############################################################################
#############################################################################
#
# 2. USER PROCEDURES
#
# Convention: 
#    (1) checks arguments extensively
#    (2) removes the interface
#    (3) calls the respective `Canon/command`
#    (4) puts the interface back
#


#
# CanonDefine
#
#    defines a tensor 
#
# Tensor is an indexed variable: 
#     e.g. R[a,b,-c,-d]
# where R is the tensor name
#       a,b are contravariant indices
#       -c,-d are covariant indices
# 
# Input: N - tensor name (no indices)
#        ni - number of indices
#        KS - list of permutations or a minus-signed permutations that
#             generates the group of symmetries of T
#
# Output: 
#   (1) an equation: CanonTable[tensor,n. of indices][0] = symmetry, base
#   (2) 0 if the symmetry cancels the tensor (e.g. {[[1,2]],[-1,[[1,2]]]})
#
Canon[CanonDefine] := proc(N::symbol, ni::nonnegint, 
    KS::{tensorsymmetry,symbol},BS::list(posint))

local strongKS,B,i;
options `Copyright 2003 by LRU Manssur and R Portugal. All rights reserved`;

if nargs<3 or nargs>4 then
     error "expecting a tensor name (1st arg), number of indices (2nd arg), \
a symmetry (3nd arg), a base (4th arg - optional)"
end if;

if nargs=4 then
     if BS=[] then B:=[1] else B := BS end if;

elif type(KS,symbol) then
     B := [$1..ni];

else 
     B := NULL  ###sort([op(indets(KS,posint))]);

end if;

forget(`Canon/Canonical/1`);

if type(KS,symbol) then

     if nargs=4 then
        error "expecting 3 arguments when the 3rd argument is a symbol"
     end if;

     if KS='Symmetric' then

        `Canon/CanonDefine`(N,ni,{seq([[i,i+1]],i=1..ni-1)},B);

     elif KS='AntiSymmetric' then

        `Canon/CanonDefine`(N,ni,{seq([-1,[[i,i+1]]],i=1..ni-1)},B);
     
     else error "expecting a set of permutations, the name Symmetric or AntiSymmetric"

     end if;

else
     if ni<max(op(indets(KS,posint))) then
        error "number of indices not compatible with symmetries"
     end if;
     strongKS := `Canon/schreier-sims`(KS,B);
     `Canon/CanonDefine`(N,ni,strongKS);
end if;
end proc:


#
# CanonicalOne
#
#    returns the canonical form of one tensor or one indexed object
#
# tensor is an indexed variable: 
#     e.g. R[a,b,-c,-d]
# where R is the tensor name
#       a,b are contravariant indices
#       -c,-d are covariant indices
# 
# Input: tensor - a tensor
#        KS - list of permutations or a minus-signed permutations that
#             generates STRONGLY the group of symmetries of tensor
#        Bs - base
#
# Output: tensor with the indices in the canonical ordering
#
Canon[CanonicalOne] := proc(T::{indexed,_TENSOR(indexed)}, 
          KS::tensorsymmetry, BS::list(nonnegint))

options `Copyright 2003 by LRU Manssur and R Portugal. All rights reserved`;

if nargs<>3 then
     error "expecting a tensor (1st arg), a symmetry (2nd arg), and a base (3rd arg)"
end if;

if KS={} and `Canon/indices`(T)[1]=[] then return T end if;
#
# `Canon/schreier-sims` is not involked
#
`Canon/CanonPrint`(
   `Canon/Canonical/Tensor`(`Canon/CanonUnPrint`(T), KS, BS ) 
                     )   
end proc:


#
# Canonical
#
#    returns the canonical form of a tensor expression
# 
# Input: expr - a tensor expression
#
# Output: expression with the indices of the tensors in the canonical ordering
#
Canon[Canonical] := proc(expr::{algebraic,list,set})

options `Copyright 2003 by LRU Manssur and R Portugal. All rights reserved`;

if nargs<>1 then 
     error "expecting 1 arg, got %1",nargs 
elif not hastype(expr,indexed) then
     return expr
elif not member(CanonMetricSymmetry,{1,0,-1}) then
     error "expecting CanonMetricSymmetry equal to 1, 0, -1, got %1",
                  CanonMetricSymmetry
end if;

`Canon/CanonPrint`(
       `Canon/Canonical`(
              `Canon/CanonUnPrint`(expr),CanonMetricSymmetry));
end proc:


#
# Puts the interface
#
Canon[CanonPrint] := proc(x) 
options `Copyright 1995-2003 by R Portugal. All rights reserved`;
`Canon/CanonPrint`(`Canon/CanonUnPrint`(x));
end proc:


#
# Removes the interface
#
Canon[CanonUnPrint] := proc()
options `Copyright 1995-2003 by R Portugal. All rights reserved`;
`Canon/CanonUnPrint`(args)
end proc:



#############################################################################
#############################################################################
#
# 3. INTERNAL GROUP PROCEDURES

#
# Warning: (1) the internal procedures do not perform type checking.
#          (2) it is expected input with no interface (_TENSOR).
#


#########################################################
# Canon/position                                        #
#                                                       #
# Input: pt - a positive integer possibly in the list   #
#        L - a list                                     #
# Output: the position of the integer in the list or 0  #
#         if not in the list                            #
#########################################################

`Canon/position` := proc(pt,L)

local pos;
options `Copyright 2003 by LRU Manssur and R Portugal. All rights reserved`;

if member(pt, L, pos) then pos
                      else 0
end if
end proc:


###########################################################
# Canon/power                                             #
#                                                         #
# Input: po - a position                                  #
#        p - a permutation or a minus-signed permutation  #
# Output: the position of the image                       #
###########################################################

`Canon/power` := proc(po,p)

local pos, c, perm;
options remember,system,`Copyright 2003 by LRU Manssur and R Portugal. All rights reserved`;

if p=[] then RETURN(po) end if;

#
# ignores sign
#

if p[1]=-1 then perm:=p[2]
           else perm:=p
end if;

#
# searches for a cycle containing the element
#

for c in perm while not member(po,c,pos) do end do;

#
# tests for trivial case
#

if not type(pos,posint) then RETURN(po)

#
# tests for end of the cycle
#

elif pos=nops(c) then RETURN(op(1,c))
                 else RETURN(op(pos+1,c))

end if

end proc:


###############################################################
# Canon/prod                                                  #
#                                                             #
# Input: p1, p2 - permutations or a minus-signed permutations #
# Output: product p1xp2, signed only if necessary             #
###############################################################

`Canon/prod` := proc(p1,p2)
options remember,system,`Copyright 2003 by LRU Manssur and R Portugal. All rights reserved`;

#
# checks trivial case
#

if p1=[] then RETURN(p2)
elif p2=[] then RETURN(p1) 
end if;

#
# output has the appropriate sign
#

if p1[1]=-1 then 
   if p2[1]=-1 then 
      `Canon/prod/nosign`(p1[2],p2[2])
   else 
      [-1,`Canon/prod/nosign`(p1[2],p2)]
   end if;

elif p2[1]=-1 then 
   [-1,`Canon/prod/nosign`(p1,p2[2])]
else 
   `Canon/prod/nosign`(p1,p2)
end if;

end proc:


#########################################
# Canon/prod/nosign                     #
#                                       #
# Input: p1, p2 - unsigned permutations #
# Output: product p1xp2                 #
#########################################

`Canon/prod/nosign` := proc(p1,p2)
local i,elts,res,rescyc,first,image;
options `Copyright 2003 by LRU Manssur and R Portugal. All rights reserved`;

#
# builds a list of nontrivial positions
#

elts := sort([op(indets({op(p1),op(p2)},integer))]);

res := NULL;
rescyc := NULL;

while elts<>[] do

   first:=elts[1];
   rescyc:=elts[1];
   image:=`Canon/power`(`Canon/power`(first,p1),p2);

   #
   # obtains a cycle, stops when it returns to first elt in cycle
   #

   while image<>first do
      rescyc := rescyc,image;
         image := `Canon/power`(`Canon/power`(image,p1),p2);
   end do;

   #
   # removes elements in this cycle from elts
   #

   elts := subs({seq(i=NULL,i=rescyc)}, elts);

   #
   # adds cycle to result
   #

   if nops([rescyc])<>1 then res := res,[rescyc] end if;

end do;

[res];

end proc:


#########################################################
# Canon/inv - inverts a permutation                     #
#                                                       #
# Input: p - a permutation or a minus-signed permutation#
# Output: the inverse of p                              #
#########################################################

`Canon/inv` := proc(p)
options remember,system,`Copyright 2003 by LRU Manssur and R Portugal. All rights reserved`;

if p=[] then 
   []

elif p[1]=-1 then 
   [-1,map(`Canon/invcyc`,p[2])]
else 
   map(`Canon/invcyc`,p)
end if;

end proc:


############################################
# Canon/invcyc - inverts a cycle           #
#                                          #
# Input: c - a cycle                       #
# Output: the inverse cycle                #
############################################

`Canon/invcyc` := proc(c)
local i;
options `Copyright 2003 by LRU Manssur and R Portugal. All rights reserved`;

[c[1], seq( c[nops(c)-i], i=0..nops(c)-2 )];

end proc:



################################################
# Canon/min - minimum with respect to a base #
################################################

`Canon/min` := proc(L,B)

local i;
options `Copyright 2003 by LRU Manssur and R Portugal. All rights reserved`;

if L={} then RETURN(NULL) end if;

for i in B do
   if member (i, L) then RETURN (i) end if;
end do;

min(op(L));

end proc:


######################################################################
# Canon/SortPermutations - sorts permutations with respect to a base #
######################################################################

`Canon/SortPermutations` := proc(L,B)
options `Copyright 2003 by LRU Manssur and R Portugal. All rights reserved`;
if nops(L)=1 then 
   L
else 
   _EnvCanonBase := B;
   sort(L,`Canon/SortPermutations/aux`);
end if;

end proc:


#
#  auxiliary procedure        
#

`Canon/SortPermutations/aux` := proc(L1,L2)

local i,p1,p2,n;
options `Copyright 2003 by LRU Manssur and R Portugal. All rights reserved`;

if L1 = [] then 'true' 
elif L2 = [] then 'false' 
elif L1=L2 then 'true'
else 

   n := max(op(_EnvCanonBase));

   for i in _EnvCanonBase do

      p1 := `Canon/power`(i,L1);
      if not member(p1,_EnvCanonBase,'p1') then 
         p1 := p1+n 
      end if;

      p2 := `Canon/power`(i,L2);
      if not member(p2,_EnvCanonBase,'p2') then 
         p2 := p2+n 
      end if;

      if p1 < p2 then RETURN('true')
      elif p1 > p2 then RETURN('false')
      end if;

   end do;

   error "base %1 not enough for ordering permutations %2,%3",
              _EnvCanonBase,L1,L2;
end if;

end proc:


##############################################################################
# Canon/orbit - calculates one orbit                                         #
#                                                                            #
# Input: p - position of a point belonging to Omega;                         #
#        K = [s1,s2,...,sn] - list of permutations or a minus-signed         #
#               permutations generating a group G acting on Omega;           #
#                                                                            #
# Output: the orbit positions p^<K>.                                         #
#                                                                            #
##############################################################################

`Canon/orbit` := proc(p,K)

local orb, new_orb, g, orb_pt, Next;
options remember,system,`Copyright 2003 by LRU Manssur and R Portugal. All rights reserved`;

orb := {p};
readlib('queue');
new_orb:= queue['new'](p);

while not queue['empty'](new_orb) do

   #
   # gets one orbit point
   #

   orb_pt := queue['dequeue'](new_orb);

   #
   # adds the action of all generators on this point to the orbit...
   #

   for g in K do

      Next := `Canon/power`(orb_pt,g);

      #
      # ...provided they are not already in orbit
      #

      if not member(Next,orb) then

         orb := orb union {Next};

         #
         # enqueues the Next points to be analised
         #

         queue['enqueue'](new_orb, Next);

      end if;
   end do;

end do;

orb

end proc:


##############################################################################
# Canon/orbit1 - calculates one orbit, schreier vector                       #
#                       and backward pointers                                #
#                                                                            #
# Input: p - position of any point belonging to Omega;                       #
#        K = [s1,s2,...,sn] - list of permutations or a minus-signed         #
#               permutations generating a group G acting on Omega;           #
#        B - a base for G;                                                   #
#        sch and back - unassigned names for the Schreier vector and         #
#                               backward pointers.                           #
#                                                                            #
# Output: the orbit positions p^<K>.                                         #
#         the Schreier vector and backward pointers components               #
#               for this orbit only                                          #
##############################################################################

`Canon/orbit1` := proc(p,K,B,sch,back)

options `Copyright 2003 by LRU Manssur and R Portugal. All rights reserved`;

`Canon/orbit2`(`Canon/min`(`Canon/orbit`(p,K),B),K,B,sch,back)

end proc:


##############################################################################
# Canon/orbit2 - calculates one orbit, schreier vector and backward          #
#       pointers, starting from the orbit representative                     #
#                                                                            #
# Input: p - an orbit representative belonging to Omega;                     #
#        K = [s1,s2,...,sn] - list of permutations or a minus-signed         #
#               permutations generating a group G acting on Omega;           #
#        B - a base for G;                                                   #
#        sch and back - unassigned names for the Schreier vector and         #
#                               backward pointers.                           #
#                                                                            #
# Output: the orbit positions p^<K>.                                         #
#         the Schreier vector and backward pointers components               #
#               for this orbit only                                          #
##############################################################################

`Canon/orbit2` := proc(p,K,B,sch,back)

local orb, new_orb, g, orb_pt, Next,rep;
options `Copyright 2003 by LRU Manssur and R Portugal. All rights reserved`;

orb := {p};
readlib('queue');
new_orb:= queue['new'](p);

while not queue['empty'](new_orb) do

   #
   # gets one orbit point
   #

   orb_pt := queue['dequeue'](new_orb);

   #
   # adds the action of all generators on this point to the orbit...
   #

   for g in K do

      Next := `Canon/power`(orb_pt,g);

      #
      # ...provided they are not already in orbit
      #

      if not member(Next,orb) then

         orb := orb union {Next};

         #
         # also builds Schreier vector and backward pointers
         #

         sch[Next] := g;
         back[Next] := orb_pt;

         #
         # enqueues the next points to be analised
         #

         queue['enqueue'](new_orb, Next);

      end if;
   end do;
end do;

rep := `Canon/min`(orb,B);
sch[rep]:=0;
back[rep]:=0;

orb

end proc:



##############################################################################
# Canon/allorbits - Calculates all the orbits, schreier vector and backward  #
#                pointers                                                    #
#                                                                            #
# Input: K = [s1,s2,...,sn] - list of permutations or a minus-signed         #
#               permutations generating a group G acting on Omega;           #
#        B - a base for G;                                                   #
#        sch and back - unassigned names for the Schreier vector and         #
#                               backward pointers.                           #
#                                                                            #
# Output: The set of orbits, the Schreier vector and the backward pointers.  #
##############################################################################

`Canon/allorbits` := proc(K, B, sch,back)

options `Copyright 2003 by LRU Manssur and R Portugal. All rights reserved`;
local orb, orbits, l, i;

sch := 'sch';
back := 'back';

#
# builds a set of nontrivial positions
#

l:= indets(K,posint);

orbits := {};

while l<>{} do

   orb := `Canon/orbit2`(`Canon/min`(l,B), K, B, sch, back);

   #
   # removes the points on the orbit from the list
   #

   l := subs({ seq(i=NULL, i=op(orb) ) }, l);

   #
   # adds the orbit to the set of orbits
   #

   orbits := orbits union {orb}

end do;

orbits;

end proc:


##############################################################################
# Canon/allorbits0 - Simply calculates all the orbits                        #
#                                                                            #
# Input: K = [s1,s2,...,sn] - list of permutations or a minus-signed         #
#               permutations generating a group G acting on Omega;           #
#                                                                            #
# Output: The set of orbits.                                                 #
##############################################################################

`Canon/allorbits0` := proc(K)

local orb, orbits, l, i;
options remember,system,`Copyright 2003 by LRU Manssur and R Portugal. All rights reserved`;

#
# builds a set of nontrivial positions
#

l:= indets(K,posint);

orbits := {};

while l<>{} do

   orb := `Canon/orbit`(l[1], K);

   #
   # removes the points on the orbit from the list
   #

   l := subs({ seq(i=NULL, i=op(orb) ) }, l);

   #
   # adds the orbit to the set of orbits
   #

   orbits := orbits union {orb}

end do;

orbits;

end proc:


###############################################################
# Canon/trace - Traces a Schreier vector                      #
#                                                             #
# Input: pos - a position                                     #
#        v - a Schreier vector                                #
#        b - the associated backward pointers                 #
#                                                             #
# Output: an element mapping the orbit representative to pos  #
###############################################################

`Canon/trace` := proc(pos, v, b)

options`Copyright 2003 by LRU Manssur and R Portugal. All rights reserved`;
local p, res;

p := pos;
res := [];

#
# Returns identity if v is not known. 
# Useful to trace singleton orbits in `Canon/doubcoset`
#

if type(v[p],indexed) then RETURN([]) end if;

while v[p]<>0 do

   res := `Canon/prod`(v[p],res);
   p := b[p];
end do;

res;

end proc:


###############################################################################
# Canon/singcoset - obtains the canonical representative of a single coset    #
#                                                                             #
# Input: g - any permutation or minus-signed permutation belonging to HxSp+2q #
#               permutations strongly generating a subgroup S of HxSp+2q;     #
#        strongK=[s1,s2,...,sn] - list of permutations or minus-signed        #
#               permutations strongly generating a subgroup S of HxSp+2q;     #
#        B - base for S with respect to strongK                               #
#        free_ind - list of positions of free indices                         #
#                                                                             #
# Output: newK - strong generating set for S that stabillizes the positions   #
#                of the free indices                                          #
#         newB - base for <newK>                                              #
#         canrep - canonical representative of the coset of S in HxSp+2q that #
#                  contains g                                                 #
###############################################################################

`Canon/singcoset` := proc(g,strongK,B,free_ind)

local canrep,newK,delta,i,sch,back,L,p,newB,omega,free,k,N,n;
options `Copyright 2003 by LRU Manssur and R Portugal. All rights reserved`;

canrep:=g;
newK:=strongK;
newB:=B;
free:=free_ind;
N:=nops(free);
n:=0;

for i in B while n<N do

   delta := [op(select(member,
                     `Canon/orbit2`(i, newK, newB, sch, back),free))];
   if delta=[] then next end if;

   L:= map(`Canon/power`, delta , canrep);
   k:=`Canon/min`(L,B);
   p:= delta[`Canon/position`(k,L)];

   omega:=`Canon/trace`(p,sch, back); 
   sch:='sch';
   back:='back';
   canrep:=`Canon/prod`(omega,canrep);

   #
   # the position of free indices changes by the action of omega
   #

   free:=map(`Canon/power`,free,`Canon/inv`(omega));

   newB:=subs(i=NULL,newB);
   newK:=remove(has,newK,i);

   n := n+1;
end do;

return canrep, newK, free;

end proc:


##############################################################################
# Canon/doubcoset - Obtains the canonical representative of a double coset   #
#                                                                            #
# Input: g - any permutation or minus-signed permutation in HxS2n            #
#         strongKS = [s1,s2,...,sn] - list of permutations or minus-signed   #
#               permutations strongly generating a subgroup S of HxS2n;      #
#        baseS - base for S with respect to strongKS                         #
#         strongKD = [d1,d2,...,dn] - list of permutations or a minus-signed #
#               permutations strongly generating the subgroup of symmetries  #
#                of dummy indices                                            #
#         baseD - base for D with respect to strongKD                        #
#         DCSize - an unused name                                            #
#                                                                            #
# Output: doubcanrep - canonical representative of the double coset          #
#                        S x g x D                                           #
##############################################################################

`Canon/doubcoset` := proc(g,strongKS,baseS,strongKD,baseD)

options remember,system,`Copyright 2003 by LRU Manssur and R Portugal. All rights reserved`;

local BS,BD,i,KS,KD,T,DeltaD,Deltab,vS,wS,vD,wD,b,alpha,images,p,j,Deltap,
        Next,s,d,L,sinv,gd,L1,F,sgd,verify_zero,DeltaS,points;

if strongKS union strongKD={} then 
        return g
end if;

#
# initializes group S
#

points := indets(strongKD,posint) union {op(baseD)}; 
###7/5/03 empty orbit problem solved

KS := strongKS;
BS := [op(baseS),op(sort([op(remove(member,points,baseS))]))];

KD := strongKD;
BD := baseD;

#
# Checks if g is in S or D
# 

if has(KS,-1) or has(KD,-1) then verify_zero := true 
                            else verify_zero := false
end if;

if not(verify_zero) and nargs<>6 and
(`Canon/isingroup`(g,KS,BS) or `Canon/isingroup`(g,KD,BD)) then

   return []

end if;


#
# initializes table T and alpha
#

T[] := [[],[]];
alpha := {[]};


for i to nops(BS)-1 do

   b := BS[i];

   DeltaD := `Canon/allorbits0`(KD);

   #
   # adds singletons to the orbits
   # 

   DeltaD := DeltaD union map(x->{x},
                remove((x,y)->has(y,x),points,DeltaD));

   DeltaS := `Canon/allorbits`(KS,BS,'vS','wS');

   Deltab := {b};

   for j in DeltaS do

      if member(b,j) then 

         Deltab := j; 
         break;
      end if;
   end do;

   images := map (`Canon/doubcoset/F1`,alpha,T,DeltaD,Deltab,g);

   p[i] := `Canon/min`(images, BS);

   BD:=[p[i], op(subs({p[i]=NULL, p[i]-1=NULL, p[i-1]=NULL}, BD )) ];

   `Canon/allorbits`(KD,BD,'vD','wD');

   for j in DeltaD do

      if member(p[i],j) then 

         Deltap := j; 
         break;
      end if;
   end do;

   for L in alpha do

      L1:=T[op(L)];

      s := L1[1];
      d := L1[2];

      Next := map( `Canon/power`, Deltab, s ) intersect 
                          map( `Canon/power` , Deltap, 
                                  `Canon/inv` (`Canon/prod`(g,d)) ) ;
      sinv := `Canon/inv`(s);
      gd := `Canon/prod`(g,d);

      for j in Next do

          T[op(L),j] := [ `Canon/prod`(
              `Canon/trace`(`Canon/power`(j,sinv),vS,wS),s),
                  `Canon/prod`(d,`Canon/inv`(
                     `Canon/trace`(`Canon/power`(j,gd),vD,wD)))];
      end do;

      T[op(L)] := evaln(T[op(L)]);
   end do;

   alpha := { indices(T) };

   #
   # verifies if the result is zero
   #

   if verify_zero then 

      F := proc(x)

      if x<>[] and x[1]=-1 then x[2] 
                           else x 
      end if
      end;

      sgd := map(`Canon/doubcoset/sgd`,alpha,T,g);

      if nops(sgd)<>nops(map(F,sgd)) then RETURN(0) end if;
   end if;

   KS := remove(has,KS,b); 

#  BS := BS[2..-1];

   KD := remove(has,KD,p[i]);
end do;

#
# if DCSize given, stores the size of the double coset in it
#
#if nargs = 6 then
#   DCSize := group[grouporder](permgroup(nops(BS), strongKS))*
#      group[grouporder](permgroup(nops(BS), strongKD))/nops(alpha);
#end if;

RETURN(`Canon/doubcoset/sgd`(alpha[1],T,g))

end proc:


#
#  F1 - auxiliary procedure
#

`Canon/doubcoset/F1` := proc(L,T,deltaD,deltab,g)

options `Copyright 2003 by LRU Manssur and R Portugal. All rights reserved`;

op( map( op , select(has,deltaD, map( `Canon/power`,deltab,
   `Canon/doubcoset/sgd`(L,T,g)    ) ) ) )
end proc:



#
#  sgd - auxiliary procedure
#

`Canon/doubcoset/sgd` := proc(L,T,g)

local L1;
options remember,system,`Copyright 2003 by LRU Manssur and R Portugal. All rights reserved`;

L1:=T[op(L)];

`Canon/prod`(`Canon/prod`(L1[1],g),L1[2]);

end proc:


##############################################################################
# Canon/SGS_dummy - Strong Generating Set for symmetry group D (dummy indices)        
#
# Input:
#
# free - positions of free indices
# q - number of pairs of dummy indices
# metric - 1, -1 or 0
#
##############################################################################

`Canon/SGS_dummy` := proc(free,q,metric)

options remember, `Copyright 2003 by LRU Manssur and R Portugal. All rights reserved`;

local BD,i;

BD := remove(member,[$1..nops(free)+2*q],free);

if metric=1 then 

   { seq( [[ BD[2*i-1],BD[2*i] ]] , i=1..q ), 
   seq( [[ BD[2*i-3], BD[2*i-1] ],
           [ BD[2*i-2], BD[2*i] ]] , i=2..q ) },BD;

elif metric=-1 then

   { seq( [-1,[[ BD[2*i-1],BD[2*i] ]]] , i=1..q ), 
   seq( [[ BD[2*i-3], BD[2*i-1] ],[ BD[2*i-2], BD[2*i] ]] , i=2..q ) },BD;

elif metric=0 then

   {seq([[BD[2*i-3],BD[2*i-1]],[BD[2*i-2],BD[2*i]]],i=2..q)},BD 

else 
   error "expecting 1,-1, or 0 as 3rd argument (metric), got %1",metric
end if;

end proc:


##############################################################################
# Canon/Canonical/Image - finds canonical configuration in image notation    #
#                                                                            #
# Input: L - a permutation in HxS2n in image notation                        #
#         KS=[s1,s2,...,sn] - list of permutations or minus-signed           #
#               permutations strongly generating a subgroup S of HxS2n;      #
#         p - number of free indices                                         #
#         q - number of PAIRS of dummy indices                               #
#         metric - specifies metric. Can be 1 (symmetric), -1 (antisymmetric)#
#                        or 0 (none)                                         #
#                                                                            #
# Output: the permutation which gives the canonical configuration            #
#                of indices, or 0, in image notation                         #
#                                                                            #
##############################################################################

`Canon/Canonical/Image` := proc(L,KS,BS,p,q,metric)

local g;
options `Copyright 2003 by LRU Manssur and R Portugal. All rights reserved`;

g := `Canon/Canonical/g`(convert(L, 'disjcyc'),KS,BS,p,q,metric);

if g=0 then 0

elif g<>[] and g[1]=-1 then

   [-1,convert(g[2],'plist',p+2*q)]
else
   [+1,convert(g,'plist',p+2*q)]
end if

end proc:


##############################################################################
# Canon/Canonical/g - Obtains the canonical configuration permutation        #
#                                                                            #
# Input: g - any permutation or minus-signed permutation in HxS2n            #
#         KS=[s1,s2,...,sn] - list of permutations or minus-signed           #
#               permutations strongly generating a subgroup S of HxS2n;      #
#        BS - base for S with respect to KS                                  #
#         p - number of free indices                                         #
#         q - number of PAIRS of dummy indices                               #
#         metric - specifies metric. Can be 1 (symmetric), -1 (antisymmetric)#
#                        or 0 (none)                                         #
#                                                                            #
# Output: the permutation which gives the canonical configuration            #
#                of indices, or 0                                            #
##############################################################################

`Canon/Canonical/g` := proc(g,KS,BS,p,q,metric)

local KD,BD,res,can_g,h,KS1,free;
options `Copyright 2003 by LRU Manssur and R Portugal. All rights reserved`;

can_g,KS1,free := `Canon/singcoset`(g,KS,BS,
                map(`Canon/power`,[$1..p],`Canon/inv`(g))   );

KD,BD:=`Canon/SGS_dummy` (free,q,metric);

h := convert([op(free) , op(BD)], 'disjcyc');

#
# Here we use BD instead of BS as the base for KS1 because of 2 reasons:
# (1) there is no natural ordering for dummy indices in terms of their names.
# (2) for Riemann monomials, non-standard bases are slower (why?).
#
res := `Canon/doubcoset`( `Canon/prod`(can_g,h), KS1, BD, KD, BD ); 

#
# To use the base given by the user, BD (3rd argument) must be replaced
# by select(member,BS,BD) as below
#
#res := `Canon/doubcoset`( `Canon/prod`(can_g,h), KS1,
#                select(member,BS,BD), KD, BD);  
#

if res=0 then 0
         else `Canon/prod`(res, `Canon/inv`(h))
end if

end proc:


##############################################################################
# Canon/isingroup - Performs membership testing                              #
#                                                                            #
# Input: g - a permutation in Sn                                             #
#         KK - a strong generating set for a subgroup G<Sn                   #
#         B - base for G with respect to KK                                  #
#                                                                            #
# Output: true if g belongs to G, false if not                               #
##############################################################################

`Canon/isingroup`:=proc(g,KK,B)

options remember,system,`Copyright 2003 by LRU Manssur and R Portugal. All rights reserved`;

local h,k,K,i,b,o,delta,v,w;

if g=[] or member(g,KK) then 
   RETURN(`true`) 
end if;

h := g;
k := nops(B);
K := KK;

for i to k do

   b := B[i];
   o := `Canon/power`(b,h);
   delta := `Canon/orbit2`(b,K,B,v,w);

   if member(o,delta) then 
      h := `Canon/prod`(h,`Canon/inv`(`Canon/trace`(o,v,w)));
   else 
      RETURN(`false`);
   end if;
        
   if i<>k then 
      K:= remove(has, K, b) 
   end if;
end do;

evalb(h=[]);

end proc:


##############################################################################
# Canon/schreier-sims - Obtains a strong generating set                      #
#                                                                            #
# Input: K - a generating set for a group G                                  #
#          optionaly, a base can be given                                    #
# Output: the base and strong generating set for G                           #
##############################################################################

`Canon/schreier-sims` := proc(K,BB)

options remember,system,`Copyright 2003 by LRU Manssur and R Portugal. All rights reserved`;

local B,base,strongK,i,Kbar,l;

if nargs=2 then
   B := BB;
else 
   B := sort([op( map(p-> 
           if p[1] = -1 then p[2][1][1] else p[1][1] end if,K) )]);
end if;

base := op(B);
strongK:=K;

for i from nops(B) to 1 by -1 do

   Kbar := remove(has,K, {op(1..i-1,B)} );
   l:= `Canon/schreier-sims/aux`([base],strongK,i,Kbar); 

   #
   # updates base and K
   #

   base:=base,op(l[2]);
   strongK:=strongK union {op(l[1])};
end do;
 
strongK,[base];

end proc:


#
#  auxiliary procedure        
#

`Canon/schreier-sims/aux`:=proc(base,strongK,i,Kbar)

options `Copyright 2003 by LRU Manssur and R Portugal. All rights reserved`;
local B,k,K,newB,newK,K_i,currentgens,oldgens,olddelta,delta,sch,back,
        gamma_,genset,s,gamma_s,trace2,trace1,g,m,j,bj,level,l;

B:=base;
#k:=nops(B);
K:=strongK;
newB:=NULL;
newK:=NULL;

K_i:=remove(has,K,{op(1..i-1,B)});

currentgens:=K_i;
oldgens:=K_i minus Kbar;
olddelta:=`Canon/orbit`(B[i],oldgens);
delta:=`Canon/orbit`(B[i],K_i);

for gamma_ in delta do

   if member(gamma_,olddelta) then genset:=Kbar
                             else genset:=currentgens;
   end if;
       
   for s in genset do

      #
      # uses current schereier vector
      #

      `Canon/allorbits`(currentgens,B,sch,back);
      gamma_s:=`Canon/power`(gamma_,s);
      trace2:=`Canon/trace`(gamma_s,sch,back);
      trace1:=`Canon/trace`(gamma_,sch,back);
      g:=`Canon/prod`(`Canon/prod`(trace1,s),`Canon/inv`(trace2));

      if not(`Canon/isingroup`(g,K,B)) then

         K:=K union {g};
         newK:=newK,g;

         k:=nops(B);

         if has(g,{op(B)}) then 
                                
            for j to k while not has(g,B[j]) do end do;
         else
            j := k+1;

            if g[1]=-1 then if g[2]=[] then
                                bj:=1 ############### check here ###############
                            else bj:= g[2][1][1];
                            end if;
                       else bj:= g[1][1];
            end if;

            B:=[op(B),bj];
            newB:=newB,bj;
         end if;

         for level from j to i+1 by -1 do

            l:=`Canon/schreier-sims/aux`(B,K,level,{g});
                
            #
            # updates B and K
            #

            for m in l[2] do

               if not member(m,B) then
 
                  B:=[op(B),m];
                  newB:=newB,m;
               end if;
            end do;
            newK:=newK,op(l[1]);
            K:=K union {op(l[1])};
         end do;
      end if;
   end do;
end do;

[[newK],[newB]];

end proc:



##############
# Type cycle #
##############

`type/cycle` := proc(x)

options `Copyright 2003 by LRU Manssur and R Portugal. All rights reserved`;

type(x,list(posint))

and nops(x)>1

#
# tests for repetitions
#

and nops(x)=nops({op(x)})

#
# tests for correct order
#

and min(op(x))=x[1]

end proc:



####################
# Type permutation #
####################

`type/permutation` := proc(x)

options `Copyright 2003 by LRU Manssur and R Portugal. All rights reserved`;

type(x, list('cycle'))

#
# tests for repetitions among cycles
#

and convert(map(nops,x),`+`)=nops({op(map(op,x))})

#
# tests for correct order
#

and sort(map2(op,1,x))=map2(op,1,x)

end proc:



###########################
# Type signed permutation #
###########################

`type/sigperm` := proc(x)

options `Copyright 2003 by LRU Manssur and R Portugal. All rights reserved`;

type(x,[1,'permutation']) or
type(x,[-1,'permutation'])

end proc:



############################
# Type general permutation #
############################

`type/genperm` := proc(x)

options `Copyright 2003 by LRU Manssur and R Portugal. All rights reserved`;

type(x,'permutation') or type(x,'sigperm')

end proc:



######################
# Type tensorsymmetry #
######################

`type/tensorsymmetry` := proc(x) #-> set of permutations/signedperm

options `Copyright 2003 by LRU Manssur and R Portugal. All rights reserved`;

type(x, set('genperm'))

end proc:





#############################################################################
#############################################################################
#
# 4. TENSOR PROCEDURES
#
# General Comments:
#
# The output may have the function _SCALAR, which has an associated 
# `print/_SCALAR` procedure. The argument of this function must be
# an indexed object with no free indices or a product of indexed 
# objects forming a scalar (no free indices) and sharing dummy indices
# (a block).
# Examples of allowed input: _SCALAR(T[a,b,-a,-b]), _SCALAR(V[a]*V[-a]),
#       _SCALAR(T[a,b,-c,-d]*T[c,d,-e,-f]*V[e]*V[f]).
#
# This input is not allowed: T[i,a]*V[-a], because has a free index (The 
# program does not check this).
#
# It was introduced to speed up calculations. To canonicalize each block
# apart is faster than to canonicalize all tensors together. For example:
# If the product R[a,b,-a,-b]*R[c,d,-c,-d]*R[e,f,g,h]*R[-e,-f,-g,-h]*
# R[i,j,k,l]*R[-i,-j,-k,-l]*R[m,n,o,p]*R[-m,-o,-n,-p] is converted to
# a permutation, the time to find the canonical representative of the 
# corresponding double coset is huge. This product is equivalent to
# _SCALAR(R[a,b,-a,-b])^2*_SCALAR(R[e,f,g,h]*R[-e,-f,-g,-h])^2*
# _SCALAR(R[m,n,o,p]*R[-m,-o,-n,-p]), which is canonicalized is less than
# a second, because each block is canonicalized separately. R is the Riemann
# tensor in this example.
#


#
# Defines a tensor - internal procedure
#
`Canon/CanonDefine` := proc(N,ni,sym,base)
global CanonTable;
options `Copyright 2003 by LRU Manssur and R Portugal. All rights reserved`;

if member([-1,[]],sym) then 
     0
else 
     CanonTable[N,ni][0] := sym,base;
     return evaln(CanonTable[N,ni][0])=sym,base;
end if;

end proc:


#
# Canonicalizes an expression - Algebraic level.
# 
# In order to remember properly, call `Canon/Canonical`(..., CanonMetricSymmetry)
#
`Canon/Canonical` := proc(expr1)

local expr,S,S1,S2;
global CanonDummyIndices,CanonFreeIndices;
options remember,system,`Copyright 2003 by LRU Manssur and R Portugal. All rights reserved`;

#
# Types list,set,`=` are allowed. Types `+`,`*`,`^` are expected.
#
if type(expr1,'{list,set,`=`}') then
     return map(`Canon/Canonical`,expr1,CanonMetricSymmetry)
elif not hastype(expr1,indexed) then
     return expr1;
end if;

#
# ``(...) is replaced by _SCALAR(...).
# ``(...) can be used in place of _SCALAR(...).
#
expr := subs(map(proc(x) 
                   if type(op(x),`*`) then
                        x='_SCALAR'(op(x))
                   else x=op(x)
                   end if
                 end proc,indets(expr1,'``(anything)')),expr1);

#
# Type function is allowed unless there are tensors in the arguments. 
# The proper way to implement function of tensors is to convert them to 
# tables beforehand, canonicalize the expression, and replace back the tables
# by functions.
#
S := select(hastype,remove(type,indets(expr,function),
                            '_SCALAR(anything)'),indexed);
if S<>{}
    then error "expecting no tensors as argument of functions, got %1",op(1,S);
end if;

#
# expands expression in order to canonicalize it recursively.
# This expansion is undesirable. It is possible to avoid this step
# using computational group theory. It was not implemented in this version.
#
expr := expand(expr);

CanonFreeIndices := `Canon/indices`(expr)[2]; # checks indices as a side effect

#
# expressions that need no canonical names for dummies
#
if type(expr,{indexed,'indexed^anything'}) then 
    return `Canon/Canonical/1`(expr, CanonMetricSymmetry) 
elif type(-expr,{indexed,'indexed^anything'}) then
    return -`Canon/Canonical/1`(-expr, CanonMetricSymmetry)
end if;

#
# Finds the all the dummy indices.
#
CanonDummyIndices := sort([op(`Canon/AllDummies`(expr))]);

#
#  no dummies
#
if nops(eval(CanonDummyIndices,1))=0 then 
   CanonDummyIndices := 'CanonDummyIndices';
   return subs(map(x->x=`Canon/Canonical/1`(x,CanonMetricSymmetry),
                 indets(expr,indexed)),expr) 
end if;

#
# changes the name of all dummy indices of scalarinvariants to
# canonical names: _PositonOfi_PositionOf-i 
#       (e.g. V[i,-i]  ->  V[_1_2,-_1_2])
# This replacement is performed at the algebraic level, so the
# following expression is simplified:
#       V[i,-i]*V[j,-j]  ->  V[_1_2,-_1_2]^2
#
expr := subs(map(x -> x=`Canon/Canonical/CanDummy`(x),
                  indets(expr,'scalarinvariant')), expr);

#
# wrap monomial blocks by the inert function _SCALAR.
# Merge expressions _SCALAR(.*.*...) to a single tensor of the form:
#    CanonTensor[T1,n1,T2,n2,...][seq of all indices]
# where T1, T2, ... are the tensor names and n1, n2, ... are
# the number of indices - they are sorted by tensorprod.
# The index names are replaced by canonical names after going 
# to the canonical position.
# This is performed at the algebraic level so it 
# simplifies expressions of the following type:
#    A[a,b]*B[-a,-b]*A[c,d]*B[-c,-d] -> 
#         CanonTensor[A,2,B,2][_1_3,_2_4,-_1_3,-_2_4]^2
#
# Attention: CanonTensor[...][...] is a scalarinvariant. It does not
# contain free indices.
#
expr := `Canon/Canonical/SplitBlocks`(
          eval(subs('_SCALAR'=
           proc(x) `Canon/Canonical/CanDummy`([op(x)]) end proc,expr)),
            {op(eval(CanonDummyIndices,1))},{op(eval(CanonFreeIndices,1))});

if expr=0 then 
    CanonDummyIndices := 'CanonDummyIndices'; 
    CanonFreeIndices := 'CanonFreeIndices';
    return 0 
end if;

if type(expr,`+`) then 
     expr := map(`Canon/Canonical/SortDummy`, expr, eval(CanonDummyIndices,1))
else 
     expr := `Canon/Canonical/SortDummy`(expr, eval(CanonDummyIndices,1))
end if:

CanonDummyIndices := 'CanonDummyIndices';
CanonFreeIndices := 'CanonFreeIndices';

#
# replaces all CanonTensors back to monomial or _SCALAR(monomial)^anything
#
if has(expr,'CanonTensor') then

     S1 := map(x->x=`Canon/Canonical/CTtoMonomial`(x),indets(expr,'CanonTensor'));

     S2 := map(x->x='_SCALAR'(
              `Canon/Canonical/CTtoMonomial`(op(1,x)))^op(2,x),
                    indets(expr,'CanonTensor^anything'));

     subs(S2, S1, expr)
     
else 
     expr

end if;

end proc:


#
# Procedure auxiliar to `Canon/Canonical`.
# Products are converted to lists or list of lists (for _FREE(product)).
# In order to remember properly, call 
# `Canon/Canonical/1`(..., CanonMetricSymmetry)
#
`Canon/Canonical/1` := proc(expr)

local symmetry,F,Terms,Terms0,Terms1;
global CanonTable;
#options remember,system,`Copyright 2003 by LRU Manssur and R Portugal. All rights reserved`;

#
# do nothing for type CanonTensor. It is already canonical.
#
if type(expr,'CanonTensor') then

     expr

#
# single tensor case
#
elif type(expr,indexed) then

     # symmetry is a sequence of the form: KS,BS
     symmetry := CanonTable[op(0,expr),nops(expr)][0];

     if nops([symmetry])=2 then
          # there is some symmetry defined
          `Canon/Canonical/Tensor`(expr,symmetry)
     else 
          # there are no defined symmetries 
          `Canon/Canonical/Tensor`(expr,{},[1])
     end if;

elif type(-expr,{indexed,'indexed^anything'}) then

     - `Canon/Canonical/1`(-expr,CanonMetricSymmetry)

elif type(expr,'_FREE(anything)') then

     if type(op(expr),`*`) then
          [`Canon/Canonical/tensorprod`([op(op(expr))])]
     else 
          [`Canon/Canonical/1`(op(expr))]
     end if

elif type(expr,list) then  # list is a product in this level

     `Canon/Canonical/tensorprod`(expr)

#
# product case: returns a list, which must be converted to type `*`.
# expr was expanded. If type expr is `*` then there is no subterm type `+`.
#
elif type(expr,`*`) then  

     Terms,Terms0 := selectremove(type,expr,
           '{scalarinvariant,scalarinvariant^anything,_FREE(anything)}');

     if type(Terms,`*`) then  

          Terms := [op(Terms)];

          Terms,Terms1 := selectremove(type,Terms,'_FREE(anything)');

          #
          # sorts _FREE(anything)'s according to the free indices. 
          #
          if eval(CanonFreeIndices,1)<>[] and nops(Terms)>1 then

               F := proc (a,b) 
                      local L;
                      L := map(proc(T)
                                 local i;
                                 global CanonFreeIndices;
                                 for i in eval(CanonFreeIndices,1) do 
                                   if has(T,i) then return i end if 
                                 end do; 
                                 return 0 
                               end proc, [a,b]);
                      evalb(L=sort(L))
                    end proc;
     
               Terms := sort(Terms, F);

          end if;

          Terms := map(proc(x) 
                         op(`Canon/Canonical/1`(x, CanonMetricSymmetry));
                       end proc,    Terms);

          #
          # Sorts Terms1 according to tensor names and number of indices
          # This order is important for dummy index renaming
          #
          F := proc(a,b)
                 local L;
                 L := map(proc(x) 
                            local T;
                            T := op(1,indets(x,indexed));
                            if type(T,'CanonTensor') then 
                                 cat('_',op(op(0,T)))
                            else evaln(_||(op(0,T))||('_')||(nops(T))) 
                            end if; 
                          end proc,  [a,b]);
                 evalb(L=sort(L))
               end proc:

          [Terms0, op(sort(Terms1,F)), op(Terms)]

     elif type(Terms,'_FREE(anything)') then  
          [Terms0,op(`Canon/Canonical/1`(Terms, CanonMetricSymmetry))]

     else  
          [Terms0,`Canon/Canonical/1`(Terms, CanonMetricSymmetry)]

     end if;

elif type(expr,`^`) then

     map(`Canon/Canonical/1`, expr, CanonMetricSymmetry)

else expr

end if;

end proc:


#
# Converts 
#        CanonTensor[T1,n1,T2,n2,...][seq of indices]
# back to 
#        T1[...]*T2[...]*...
#
# Example:
#          Input:  CanonTensor[A,1,B,1][_1_2,-_1_2]
#          Output: A[a]*B[-a]
#
`Canon/Canonical/CTtoMonomial` := proc(x)
options remember,system,`Copyright 2003 by R Portugal. All rights reserved`;
    `Canon/Canonical/CTtoMonomial/1`([op(op(0,x))],[op(x)])
end proc:

`Canon/Canonical/CTtoMonomial/1` := proc(L,ind)
options `Copyright 2003 by LRU Manssur and R Portugal. All rights reserved`;
if L=[] then
     1
else 
     L[1][op(1..L[2],ind)]*procname(L[3..-1],ind[L[2]+1..-1])
end if;
end proc:


#
# Replaces dummy indices of scalarinvariants or scalar monomials
# by canonical names after full canonicalization. A scalar monomial
# is converted to CanonTensor[...][indices]. The indices have the
# form:
#      _PositonOfi_PositionOf-i 
# e.g. 
#   (1)   Input:  V[i,-i]
#         Output: V[_1_2,-_1_2]
#
#   (2)   Input:  A[a]*B[-a]  (or [A[a],B[-a]])
#         Output: CanonTensor[A,1,B,1][_1_2,-_1_2]
#
`Canon/Canonical/CanDummy` := proc(TT)
options `Copyright 2003 by LRU Manssur and R Portugal. All rights reserved`;
local L,i,SS,T,Sign;
global CanonDummyIndices;

if type(TT,'scalarinvariant') then
      T := `Canon/Canonical/1`(TT, CanonMetricSymmetry);
      if T=0 then
         return 0
      elif type(-T,indexed) then
         Sign := -1;
         T := -T;
      else Sign := 1;
      end if;
 
     L := [op(T)];

     SS := {seq(i=evaln(_||(`Canon/position`(i,L))||
                ('_')||(`Canon/position`(-i,L))), 
                      i=`Canon/indices`(T)[1])};

     CanonDummyIndices := [op(eval(CanonDummyIndices,1)),
                       op(map(rhs,SS) minus {op(eval(CanonDummyIndices,1))})];

     T := subs(SS,T);

     # Remember the result with canonical indices
     `Canon/Canonical/1`(T,CanonMetricSymmetry) := T;

     return Sign*T;
   
elif type(TT,'{`*`,list}') then

     T := `Canon/Canonical/1`(TT, CanonMetricSymmetry);

     if member(0,T) then
          return 0
     else 
          T,Sign := selectremove(type,T,indexed);
     end if;

     L := map(op,T);

     SS := {seq(i=evaln(_||(`Canon/position`(i,L))||
                ('_')||(`Canon/position`(-i,L))), 
                      i=`Canon/AllDummies`(convert(TT,`*`)))};

     CanonDummyIndices := [op(eval(CanonDummyIndices,1)),
                       op(map(rhs,SS) minus {op(eval(CanonDummyIndices,1))})];

     convert(Sign,`*`)*CanonTensor[op(map(x->(op(0,x),nops(x)),
                                         T))][op(subs(SS,map(op,T)))];
   
else error "expecting type scalarinvariant, `*` or, list , got %1",TT
end if;
end proc:


#
# Split the product in blocks of terms that share dummy indices.
# Canonicalize and convert to CanonTensor the blocks with no free 
# indices with 2 or more tensors.
# Wrap blocks with free indices (and at least a pair of dummy indices)
# with the inert function _FREE.
# This is very important for performance.
# E.g.
#      A[a]*B[-a]*C[b]*D[-b] ->  CanonTensor[A,1,B,1][_1_2,-_1_2]^2
#      T[i,a,-a]    -> _FREE(T[i,a,-a])  (no dummy renaming)
#      V[i,a]*W[-a] -> _FREE(V[i,a]*W[-a])
#
`Canon/Canonical/SplitBlocks` := proc(expr1,Dummy,Free)

options `Copyright 2003 by R Portugal. All rights reserved`;
local i,j,L,F,F1,S,S1,expr;

if type(expr1,'{`+`,set}') then

     map(procname,expr1,Dummy,Free);

elif type(expr1,`*`) then

     L,expr := selectremove(x->type(x,indexed) and 
                   not type(x,'{scalarinvariant}'), expr1);

     if type(L,`*`) then
        L := [op(L)];

        F := proc(x,j,i) if has(x,j) then i else NULL end if end proc:

        S:=`Canon/tools/joinsets`({seq({seq(F(L[i],j,i),
                              i=1..nops(L))},j=Dummy)});

        S,S1 := selectremove(x->nops(x)>1,S);

        S1 := map(op,S1) union ({$1..nops(L)} minus map(op,S));

        F := proc(x) 
               if has(x,Free) then 
                     '_FREE'(x) 
               else `Canon/Canonical/CanDummy`([op(x)]) 
               end if 
             end proc;

        F1 := proc(x)
                if has(x,Dummy) then 
                     '_FREE'(x)
                else
                     `Canon/Canonical/1`(x, CanonMetricSymmetry)
                end if
              end proc;

        expr*mul(F1(L[i]),i=S1)*mul(F(mul(L[j],j=i)),i=S);

     else expr1
     end if; 
else expr1
end if;
end proc:


#
# Canonicalizes expr and sorts the dummy indices following the
# order given by CanonDummyIndices. This procedure replaces the
# dummy canonical names by the dummy names used in the input of
# the procedure Canonical.
#
`Canon/Canonical/SortDummy` := proc(expr,Dummies)
options `Copyright 2003 by LRU Manssur and R Portugal. All rights reserved`;
local x;

x := `Canon/Canonical/1`(expr, CanonMetricSymmetry);

if not type(x,list) then x:=[x] end if;

convert(map(proc(x) 
              if type(x,list) then op(x)
              else x
              end if;
            end proc, `Canon/Canonical/SortDummy/1`(x, Dummies)),`*`)

end proc:

#
# finds the sequence of all indices of a tensor, tensor^anything, or
# a monomial (in the form of a list), replaces all indices using
# the entries of CanonDummyIndices.
#
`Canon/Canonical/SortDummy/1` := proc(LL,Dummies)
options `Copyright 2003 by LRU Manssur and R Portugal. All rights reserved`;
local i,L;
global CanonDummyIndices;

if LL=[] then []
else 
     L := select(member,[`Canon/Canonical/SortDummy/2`(LL[1])],
                  eval(CanonDummyIndices,1));

     [subs({seq(L[i]=Dummies[i],i=1..nops(L))}, LL[1]),
               op(procname(LL[2..-1],Dummies[nops(L)+1..-1]))];

end if;

end proc:

`Canon/Canonical/SortDummy/2` := proc(x)
options `Copyright 2003 by LRU Manssur and R Portugal. All rights reserved`;

if type(x,indexed) then
     op(x)
elif type(-x,indexed) then
     op(-x)
elif type(x,'indexed^anything') then
     op(op(1,x))
elif type(-x,'indexed^anything') then
     op(op(1,-x))
elif type(x,list) then
     op(map(`Canon/Canonical/SortDummy/2`,x))
else 
     NULL
end if

end proc:


#
# canonicalizes single tensors
#
`Canon/Canonical/Tensor` := proc(T,KS,BS)

options `Copyright 2003 by R Portugal. All rights reserved`;
local  i,L,ind,p,q,can_L,free,dummy;

# 
# free and p store information on the free indices 
# dummy and q store information on the dummy indices
#
ind := `Canon/indices`(T);
p := nops(ind[2]);
q := nops(ind[1]);

if q=0 and KS={} then return(T) end if;

free := `Canon/tools/sort`(ind[2]);
dummy := sort(ind[1]);

#
# L stores the positions of indices
#
L := subs({seq(free[i]=i,i=1..p)},
        {seq(-dummy[i]=p+2*i,i=1..q)},
                {seq(dummy[i]=p+2*i-1,i=1..q)},[op(T)]);
#
# calls computational group algorithms
#
can_L := `Canon/Canonical/Image`(L,KS,BS,p,q,CanonMetricSymmetry);

if can_L=0 then RETURN(0) end if;

#
# restores user-level indices and outputs the result
#
L := subs({seq(i=free[i],i=1..p)},
        {seq(p+2*i=-dummy[i],i=1..q)},
                {seq(p+2*i-1=dummy[i],i=1..q)},can_L[2]);

can_L[1]*op(0,T)[op(L)];

end proc:


#
# `Canon/Canonical/tensorprod` 
#            canonicalize a block of tensors (tensor product)
#
# INPUT: L is a list of at least 2 tensors sharing dummy indices. 
# OUTPUT: a sequence with the tensor indices in the canonical ordering.
#         Some of the operands can be -1.
#         The output can be [0]
#
`Canon/Canonical/tensorprod` := proc(Tensors1)

local T,Tensors,N_Tensors,TensorNames,L_N_I,NewNames,SameNames,s_subsop1,
s_subsop2,i,pos,N,p,sym_T,MergedTensors,N_MergedTensors,L_all_ind,L_N_I1,
F,L_displ,merged_sym,merged_base,Merged_T,TensorSign,S,S_inv,j,jj;

global CanonTable;

options `Copyright 1997-2003 by R Portugal All rights reserved`;

if type(Tensors1,'_FREE(anything)') then
     Tensors := [op(op(Tensors1))];  
else 
     Tensors := Tensors1;
end if;

#
# sorts Tensors according to tensor names and number of indices
#
Tensors := sort(Tensors,proc(a,b)
                          local L;
                          L := map(proc(x) 
                                     if type(x,indexed) then
                                        evaln(_||(op(0,x))||('_')||(nops(x)))
                                     else 
                                        0
                                     end if
                                   end proc,  [a,b]);
                          evalb(L=sort(L))
                        end proc);

N_Tensors := nops(Tensors);
TensorNames  := map2(op,0,Tensors);
L_N_I        := map(nops,Tensors); # List of Number of Indices 
NewNames   := [seq(evaln(Canon_||(TensorNames[i])||('_')||
                      (L_N_I[i])), i=1..N_Tensors)];
SameNames := `Canon/selectrepeated`(NewNames);

#
# Merging equal tensors (same name and same n. of indices)
#
s_subsop1   := NULL;
s_subsop2   := NULL;

for i in SameNames do
  pos := [`Canon/positions`(i,NewNames)];
  N   := nops(pos);
  p   := L_N_I[pos[1]];

  if not `Canon/has_sym`(Tensors[pos[1]],'sym_T') then 
       sym_T := {},[1];
  end if;

  `Canon/CanonDefine`(i,p*N,`Canon/Canonical/mksym`(p,N,sym_T));

  s_subsop1 := s_subsop1, seq(j=NULL,j=pos[2..-1]),
           pos[1]=i[seq(op(Tensors[jj]),jj=pos)];
  s_subsop2 := s_subsop2, seq(j=NULL,j=pos[2..-1]),
           pos[1]=i;
end do;

MergedTensors  := subsop(s_subsop1, Tensors);
N_MergedTensors := nops(MergedTensors);
NewNames := subsop(s_subsop2, NewNames);

#
# Full reduction: all tensors merge to form one tensor: Merged_T
#
L_all_ind := map(op,MergedTensors);
L_N_I1 := map(nops,MergedTensors);

#
# calculates the displacement
#
F := proc(L)
 options remember,system;
  if L=[] then 
       0
  else
       F(L[1..-2]),op(-1,[F(L[1..-2])])+L[-1]
  end if
end proc:

L_displ := [F(L_N_I1[1..-2])];

F := proc(x) if type(x,set) then x else {} end if; end proc:

merged_sym := {seq(op(`Canon/Canonical/sym_displ`(L_displ[i],
                F(CanonTable[op(0,MergedTensors[i]),L_N_I1[i]][0][1]))),
                   i=1..N_MergedTensors)};

F := proc(x) if type(x,list) then x else [] end if; end proc:

merged_base := [seq(op(`Canon/Canonical/sym_displ`(L_displ[i],
                F(CanonTable[op(0,MergedTensors[i]),L_N_I1[i]][0][2]))),
                   i=1..N_MergedTensors)];

T := 'T';
Merged_T := `Canon/Canonical/Tensor`(T[op(L_all_ind)],merged_sym,merged_base);

#
# Clear the symmetries of the tensors defined above
#
for i to N_MergedTensors do
   if not member(op(0,MergedTensors[i]),TensorNames) then

#
# The next command is the proper command to unassing. It is not
# working. This seems to be a bug in unassign.
#
#       unassign(evaln(CanonTable[op(0,MergedTensors[i]),L_N_I1[i]]));

   CanonTable := table(remove(has,op(eval(CanonTable)),
                                    op(0,MergedTensors[i])));

   end if
end do;

#
# recovering the full reduction
#
if Merged_T=0 then 
   return [0] 
elif type(Merged_T,indexed) then
   TensorSign := 1;
elif type(-Merged_T,indexed) then
   Merged_T := -Merged_T;
   TensorSign := -1;
else error "internal error: expecting 0, indexed, \
or minus indexed, got %1",Merged_T
end if;

#
# replace temporarily -i by the local synonym i
#
S := map(y->y=`tools/gensym`(-y),select(x->type(-x,symbol),{op(L_all_ind)}));
S_inv := map(x->rhs(x)=lhs(x),S);

Tensors := subs(S,Tensors);

S := subs(S,{seq(L_all_ind[i]=op(i,Merged_T),i=1..nops(L_all_ind))});

[TensorSign, op(subs(S,S_inv,Tensors))];

end proc:



#
# Returns true if a tensor has symmetry and false otherwise.
# If a 2nd argument is given, the symmetry is assigned to it.
#
`Canon/has_sym` := proc(expr,p)
local n,symmetry;
global CanonTable;
options `Copyright 1997 by R Portugal All rights reserved`;
n:=nops(expr);
if n<2 
  then false
  else symmetry := CanonTable[op(0,expr),n][0];
       if nops([symmetry])=2 then
            if nargs=2 then p := symmetry end if;
            true
       else 
            false
       end if
end if
end proc:


#
# `Canon/Canonical/mksym`
#
#      creates symmetry for product of equal tensors
#
# Input: n - number of indices of each tensor
#        N - number of equal tensors
#        sym_T - symmetry of the original tensor
# 
# Output: a strong generating set of the symmetry of the product of tensors 
#
`Canon/Canonical/mksym` := proc(n,N,sym,Base)

options `Copyright 2003 by LRU Manssur and R Portugal. All rights reserved`;

local i,j,base;

if Base=[] then 
     base := [1]
else
     base := Base
end if;

{ seq(op(`Canon/Canonical/sym_displ`(j*n,sym)),j=0..N-1),
     seq([seq([i+j*n,n+i+j*n],i=1..n)],j=0..N-2)  },
[ seq(op(`Canon/Canonical/sym_displ`(j*n,base)),j=0..N-1)  ]

end proc:


#
# `Canon/Canonical/sym_displ`
#
#      add i to the symmetry
#
# Input: i - non negative integer
#        sym_T - symmetry
# 
# Output: add i to all permutations 
#
`Canon/Canonical/sym_displ` := proc(i,sym_T)

options `Copyright 2003 by LRU Manssur and R Portugal. All rights reserved`;

subs(map(proc(x) x=x+i end proc,indets(sym_T,posint)),sym_T)

end proc:


#
# `Canon/indices`
#
#      returns the indices of a tensor or tensor expression
#
# Input: expr
# Output: [ [dummy indices], [free indices], [numerical indices] ]
#
# Attention: The output does not contain all indices necessarily, dummy and 
# numerical indices can be missing in expr of type "+". Eg. T[i,-i,1]+T[j,-j,2]
#
`Canon/indices` := proc(expr) 

local nameind,repeated,ind,numind,s; 
options remember,system,`Copyright 1995-2003 by R Portugal. All rights reserved`;

if nargs=0 or not hastype(expr,indexed) then RETURN([[],[],[]]) end if;

ind:=[`Canon/indices/rec`(expr)];

nameind:=select(type,ind,'nameindex');
numind :=select(type,ind,'numericindex');

#
# Next line does not allow the presence of numerical indices
# in the Canon package.
#
if nops(numind)>0 then 
  error "expecting no numerical indices, but got %1",numind 
end if;

if nops(numind)+nops(nameind)<>nops(ind) then 
  ERROR(`index of wrong type`)
end if;

if nops({op(nameind)})<>nops(nameind) then 
  ERROR(`wrong choice of indices`) 
end if;

repeated := select(type,subs(0=NULL,subs(
              map(proc(x) x=0 end,{op(convert(nameind,`+`))}),
                nameind)),'symbol');

s := subs(0=NULL,subs( map(proc(x) x=0 end,repeated) ,nameind));

[repeated, s, numind]

end proc:


`Canon/indices/rec` := proc(expr)  
local i,ind,ind1,op0,n,s;
#options remember,system;
options `Copyright 1997-2003 by R Portugal. All rights reserved`;
if type(expr,'indexed') 
   then op(expr)
elif type(expr,'function')
   then op0:=op(0,expr);
        if op0='_SCALAR'
          then procname(op(expr))
        elif member(op0,'{diff,Diff}')
          then ind:=`Canon/indices`(op(1,expr));
               ind1:=procname(op(2,expr));
               if ind1=NULL 
                 then  op(op(1,ind)),op(map(proc(x) -x end,op(1,ind))),
                             op(op(2,ind)),op(op(3,ind))
               elif member(op0,'{diff,Diff}') and
                    (nops([ind1])>1 or type(ind1,'covariantindex') or 
                     (type(-ind1,'symbol') and has(ind,-ind1)))
                 then ERROR(`wrong choice of indices in derivative`) 
               else  op(op(1,ind)),op(map(proc(x) -x end,op(1,ind))),
                             op(op(2,ind)),op(op(3,ind)),-ind1
               end if
        elif type(op0,'function') 
          then s:=op(0,op0);
               if s='D' then ind:=`Canon/indices`(op(op0)); 
                             op(op(2,ind)),op(op(3,ind))
               elif type(s,'indexed') and op(0,s)='D'
                 then if hastype(s,'covariantindex')
                        then ERROR(`wrong choice of indices in D operator`)
                      end if;
                      ind:=`Canon/indices`(op(op0));
                      op(op(2,ind)),op(op(3,ind)),op(map(x->-x,[op(ind)]))
               elif not hastype(s,'indexed')
                 then NULL
               else ERROR(`NOT IMPLEMENTED`)
               end if 
        elif nops(expr)=1
           then ind:=seq(op(i),i=op(2..3,`Canon/indices`(op(expr))));
                if ind=NULL then NULL
                else userinfo(1,'Canon',`assuming that function `||op0||
                      ` is linear in its argument`);
                     ind
                end if
        elif nops(expr) > 1
          then if hastype(expr,'constanttensor')
               then ind:=seq(op(i),i=op(2..3,`Canon/indices`(op(1,expr))));
                    if ind=NULL then NULL
                    else userinfo(1,'Canon',`assuming that function `||op0||
                                    ` has tensors in its 1st argument only`);
                         ind
                    end if
               else NULL
               end if
        else NULL         
        end if
elif type(expr,`*`) 
   then ind1:=seq(`Canon/indices`(i),i=expr);
        ind:=seq(op(op(2,i)),i=ind1);
        if {seq(op(op(1,i)),i=ind1)} intersect {seq(`Canon/abs`(i),i=ind)} <> {}
          then ERROR(`dummy index equal to free index`) 
        end if;
        ind,seq(op(op(3,i)),i=ind1)
elif type(expr,`+`)
   then ind:=map(proc(x) op(2..3,`Canon/indices`(x)) end,[op(expr)]);
        n:=nops(ind)/2;
        if not(nops({seq({op(op(2*i-1,ind))},i=1..n)})=1   )
        then `Canon/CanonPrint`(expr);
             ERROR(`free indices do not match or summing tensors of different type`) 
        end if;
        procname(op(1,expr))
elif type(expr,`^`) 
   then ind:=map(`Canon/indices`,[op(expr)]);
        if op(2,op(1,ind))=[] and op(2,op(2,ind))=[]
          then NULL
        else  `Canon/CanonPrint`(expr);
              ERROR(`free indices in power`)
        end if
else NULL
end if 
end proc:


#
# finds all dummy indices of an expression
#
`Canon/AllDummies` := proc(expr)
options `Copyright 2003 by R Portugal. All rights reserved`;
    map(x->op(`Canon/indices`(x)[1]),indets(expr,indexed)) union 
      map(x->op(`Canon/indices`(op(x))[1]),indets(expr,'_SCALAR(anything)')) 
        union {`Canon/AllDummies/1`(expr)}
end proc:

`Canon/AllDummies/1` := proc(expr)
options `Copyright 2003 by R Portugal. All rights reserved`;
local i;
if type(expr,'{`+`,set,list}') then
     seq(`Canon/AllDummies/1`(i),i=expr)
elif type(expr,`*`) then 
     op(`Canon/indices`(expr)[1])
else NULL
end if;
end proc:




#############################################################################
#############################################################################
#
# 5. TYPES (tensor)


`type/constanttensor` := proc(x)
options remember,system,`Copyright 1995-2003 by R. Portugal. All right reserved.`;
type(x,'indexed') and type(op(0,x),'symbol') and
type([op(x)],list('{nameindex,posint}'))
end proc:

`type/scalarinvariant` := proc(x)
options remember,system,`Copyright 1995-2003 by R. Portugal. All right reserved.`;
type(x,indexed) and map(op,`Canon/indices`(x)[2..3])=[]
end proc:

`type/CanonTensor` := proc(x)
options remember,system,`Copyright 2003 by LRU Manssur and R. Portugal. All right reserved.`;
type(x,indexed) and type(op(0,x),indexed) and op([0,0],x)='CanonTensor'
end proc:

`type/contravarianttensor` := proc(x)
options `Copyright 1997 by R. Portugal. All right reserved.`;
type(x,'constanttensor') and type([op(x)],list('contravariantindex'))
end proc:

`type/covarianttensor` := proc(x) 
options `Copyright 1997 by R. Portugal. All right reserved.`;
type(x,'constanttensor') and type([op(x)],list('covariantindex'))
end proc:

`type/nameindex` := proc(x) 
options `Copyright 1995-2003 by R Portugal. All rights reserved`;
type(x,'algebraic') and not type(x,'numericindex') and 
(type(x,'symbol') or type(-x,'symbol')) 
end proc:

`type/contravariantnameindex` := proc(x) 
options `Copyright 1995-2003 by R Portugal. All rights reserved`;
type(x,'symbol') 
end proc:

`type/covariantnameindex` := proc(x) 
options `Copyright 1995-2003 by R Portugal. All rights reserved`;
type(x,algebraic) and type(-x,'symbol')  
end proc:

`type/contravariantindex` := proc(x) 
options `Copyright 1995-2003 by R Portugal. All rights reserved`;
type(x,'{symbol,contravariantnumericindex}') 
end proc:

`type/covariantindex` := proc(x) 
options `Copyright 1995-2003 by R Portugal. All rights reserved`;
type(x,algebraic) and type(-x,'{symbol,contravariantnumericindex}')  
end proc:

`type/numericindex` := proc(x)
options `Copyright 1995-2003 by R Portugal. All rights reserved`;
if not type(x,'algebraic') then 'false'
elif not type(x,integer) then 'false'
elif x=0
  then 'false'
else true 
end if
end proc:

`type/contravariantnumericindex` := proc(x)
options `Copyright 1995-2003 by R Portugal. All rights reserved`;
if not type(x,'algebraic') then 'false'
elif not type(x,nonnegint) then 'false'
elif x=0
  then 'false'
else true 
end if
end proc:

`type/covariantnumericindex` := proc(x) 
options `Copyright 1995-2003 by R Portugal. All rights reserved`;
if not type(x,'algebraic') then 'false'
elif not type(-x,nonnegint) then 'false'
elif x=0
  then 'false'
else 'true' 
end if
end proc:



#############################################################################
#############################################################################
#
# 6. INTERFACE


`print/_SCALAR` := proc(x) ``(x) end:


`Canon/CanonPrint` := proc(expr)
local op0;
options `Copyright 1995-2003 by R Portugal. All rights reserved`;
if type(expr,'indexed') then 
     _TENSOR(expr)
elif type(expr,'function') then 
     op0:=op(0,expr);
     if op0='_SCALAR' then
          '_SCALAR'(procname(op(expr)))
     elif op0='diff' or op0='Diff' then 
          op0(op(map(procname,[op(expr)])))
     elif type(op0,'indexed') then 
          _TENSOR(expr)
     elif has(op0,'D') then 
          op(0,op0)(procname(op(op0)))(op(map(procname,[op(expr)])))
     else op0(op(map(procname,[op(expr)])))
     end if
elif type(expr,{`+`,`*`,`=`,list,set}) then 
     map(procname,expr)
elif type(expr,`^`) then 
     if type(op(1,expr),'constanttensor') and nops(op(1,expr))<>0 and 
                   op(2,expr)<>-1 then 
          '_SCALAR'(procname(op(1,expr)))^procname(op(2,expr))
     else map(procname,expr)
     end if
else expr
end if
end proc:


_TENSOR := proc()
options `Copyright 1997-2003 by R Portugal. All rights reserved`;
if type(args[1],`*`) then 
       map('_TENSOR',args[1],args[2..-1])
  else '_TENSOR'(args)
end if
end proc:


`print/_TENSOR` := proc(x)
options remember,system,`Copyright 1995 by R. Portugal. All right reserved.`;
if type(x,'indexed') then
     if nops(x)=0 then 
            op(0,x)
     else convert([readlib('`tools/gensym`')(op(0,x)),
                                   op(map(`_TENSOR/F` ,[op(x)]))],`*`)
     end if
elif type(x,'function') then 
     if type(op(0,x),'indexed') then 
              convert([readlib('`tools/gensym`')(op(0,op(0,x))),
                 op(map(`_TENSOR/F` ,[op(op(0,x))])),
                    readlib('`tools/gensym`')(``)(
                       op(map(procname,[op(x)])))],`*`)
     else x
     end if
else x
end if
end proc:


`_TENSOR/F` := proc(x)
options `Copyright 1997-2003 by R Portugal. All rights reserved`;
#
# Maple 9 requires ` ` instead of `` for representing upper indices 
#
   if x=1 then readlib('`tools/gensym`')(` `)^`1`
   elif x=-1 then readlib('`tools/gensym`')(``)[`1`]
   elif type(x,'contravariantindex') then readlib('`tools/gensym`')(` `)^x
   elif type(x,'covariantindex') then readlib('`tools/gensym`')(``)[-x]
   else error "%1 is not an index of correct type",x
   end if
end proc:


`Canon/CanonUnPrint` := proc(expr1)
local expr;
options `Copyright 1997-2003 by R Portugal. All rights reserved`;
expr := subs(map(x->x=op(x),indets(expr1,'_TENSOR(anything)')),expr1);
subs(map(x->x=op(x),indets(expr,'_SCALAR(scalarinvariant)')),expr);
end proc:



#############################################################################
#############################################################################
#
# 7. GENERAL TOOLS


#
# Canon/positions - positions of an element occuring in a list
#
`Canon/positions` := proc(i,L)

options `Copyright 1997-2003 by R Portugal. All rights reserved`;
local p,j;

if member(i,L,'p')
        then p,seq(p+j,j=procname(i,L[p+1..-1]))
        else NULL
end if

end proc:


#
# `Canon/selectrepeated` 
#
#         selects the repeated elements of a list preserving the order
#
# Example:  [A,A,A,B,C,C,D,E] -> [A,C]
#
`Canon/selectrepeated` := proc(L)
local f1,L1;
options `Copyright 1997-2003 by R Portugal. All rights reserved`;
f1 := proc(x,L) local pos; member(x,L,pos); pos=NULL end;
L1 := {op(subsop(op(map(f1,{op(L)},L)),L))};
map(op,sort([op(map(`Canon/position`,L1,L))]),L);
end proc:


#
# Sorts L using lexicographical order for contravariant indices followed
# by the covariant indices:
# [ a, b, c, ...,  -d, -e, -f, ...] 
#
# Numeric indices is not allowed.
#
`Canon/tools/sort` := proc(L) 
local L_contra,L_co;
options `Copyright 1997-2003 by R Portugal. All rights reserved`;

L_contra,L_co := selectremove(type,L,'contravariantnameindex');

[op(sort(L_contra,'lexorder')),
   op(map(x->-x,sort(map(x->-x,L_co),'lexorder')))]

end proc:


#
# input: a list, a set or an index
#
`Canon/abs` := proc(x) 
options `Copyright 1995-2003 by R Portugal. All rights reserved`;
if type(x,{'list,set'}) then map(procname,x)
elif type(x,{'symbol','posint'}) then x
elif type(-x,{'symbol','posint'}) then -x
elif x=0
  then error "zero cannot be used as index"
else error "wrong index type %1",x 
end if
end proc:


#
# Join sets that have at least one element in common.
# E.g.: {{6},{1,2},{2,3},{4,5},{6,7},{4,7,8}}  ->
#                      {{4, 5, 6, 7, 8}, {1, 2, 3}}
#
`Canon/tools/joinsets` := proc(S)
local S1,S2,s;
options `Copyright 1997-2003 by R Portugal. All rights reserved`;

if nops(S)<2
  then S
else S1 := procname(S[2..-1]);
     s  := S[1];
     S2  := select(has,S1,s);
     {s union map(op,S2)} union remove(member,S1,S2)
end if;
end proc:


##### THE END #####

